/*
    Copyright (C) 2002 Timo Sirainen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "lib.h"
#include "iobuffer.h"
#include "imap-message-send.h"

#ifdef HAVE_SYS_SENDFILE_H

#include <unistd.h>
#include <sys/sendfile.h>

static int message_send_sendfile(int out_fd, int in_fd, size_t size)
{
	off_t offset;
	ssize_t ret;

	offset = lseek(in_fd, 0, SEEK_CUR);
	if (offset == (off_t)-1)
		return FALSE;

	while (size > 0) {
		/* FIXME: create sub-ioloop here like it's done with
		   blocking iobuffer sends .. or maybe move this thing
		   entirely into lib/iobuffer.c?

		   anyway, the problem here is that a) we never timeout,
		   b) we're nonblocking which means this loop eats cpu. */
		ret = sendfile(out_fd, in_fd, &offset, size);
		if (ret < 0)
			break;

		size -= ret;
	}

	return size == 0;
}
#endif

int imap_message_send(IOBuffer *outbuf, const char *msg, int msg_fd,
		      MessageSize *size, off_t virtual_skip,
		      size_t max_virtual_size)
{
	const char *msg_start, *msg_end, *cr;
	int len;

	if (size->physical_size == 0)
		return TRUE;

	if (size->physical_size == size->virtual_size) {
		/* no need to kludge with CRs, use sendfile() */
#ifdef HAVE_SYS_SENDFILE_H
		if (msg_fd != -1) {
			size_t send_size;

			if (virtual_skip != 0) {
				if (lseek(msg_fd, virtual_skip,
					  SEEK_CUR) == (off_t)-1)
					return FALSE;
			}

			io_buffer_send_flush(outbuf);
			send_size = size->physical_size - virtual_skip;
			return message_send_sendfile(outbuf->handle,
						     msg_fd, send_size);
		}
#endif
		return io_buffer_send(outbuf, msg + virtual_skip,
				      size->physical_size - virtual_skip) > 0;
	}

	msg_start = msg;
	msg_end = msg + size->physical_size;

	/* first do the virtual skip - FIXME: <..\r><\n..> skipping! */
	if (virtual_skip > 0) {
		cr = NULL;
		while (msg != msg_end && virtual_skip > 0) {
			if (*msg == '\r')
				cr = msg;
			else if (*msg == '\n') {
				if (cr != msg-1) {
					if (--virtual_skip == 0) {
						/* FIXME: cr thingy */
					}
				}
			}

			msg++;
			virtual_skip--;
		}

		msg_start = msg;
	}

	/* go through the message data and insert CRs where needed.  */
	cr = NULL;
	while (msg != msg_end) {
		if (*msg == '\r')
			cr = msg;
		else if (*msg == '\n' && cr != msg-1) {
			len = (int) (msg - msg_start);
			if (max_virtual_size != 0 && max_virtual_size <= len) {
				/* reached max. size limit */
				return io_buffer_send(outbuf, msg_start,
						      max_virtual_size) > 0;
			}

			if (io_buffer_send(outbuf, msg_start, len) <= 0)
				return FALSE;

			if (io_buffer_send(outbuf, "\r", 1) <= 0)
				return FALSE;

			/* update max. size */
			if (max_virtual_size == len+1)
				return TRUE;
			max_virtual_size -= len+1;

			msg_start = msg;
		}
		msg++;
	}

	/* send the rest */
	len = (int) (msg - msg_start);
	if (max_virtual_size != 0 && max_virtual_size < len)
		len = max_virtual_size;
	return io_buffer_send(outbuf, msg_start, len) > 0;
}
