/*
    Copyright (C) 2002 Timo Sirainen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "lib.h"
#include "maildir-index.h"
#include "mail-index-util.h"

static MailIndex maildir_index;

const char *maildir_filename_set_flags(const char *fname, MailFlags flags)
{
	/* FIXME: we should leave the unknown flags */
	const char *info;
	char flags_str[10], *p;

	/* convert flags to ASCII - must be sorted */
	p = flags_str;
	if (flags & MAIL_DRAFT)
		*p++ = 'D';
	if (flags & MAIL_FLAGGED)
		*p++ = 'F';
	if (flags & MAIL_ANSWERED)
		*p++ = 'R';
	if (flags & MAIL_SEEN)
		*p++ = 'S';
	if (flags & MAIL_DELETED)
		*p++ = 'T';
	*p = '\0';

	/* remove the old :info from file name */
	info = strrchr(fname, ':');
	if (info != NULL && strrchr(fname, '/') > info)
		info = NULL;

	if (info != NULL)
		fname = t_strndup(fname, (int) (info-fname));

	return t_strconcat(fname, ":2,", flags_str, NULL);
}

MailIndex *maildir_index_alloc(const char *dir)
{
	MailIndex *index;
	int len;

	i_assert(dir != NULL);

	index = i_new(MailIndex, 1);
	memcpy(index, &maildir_index, sizeof(MailIndex));

	index->fd = -1;
	index->dir = i_strdup(dir);

	len = strlen(index->dir);
	if (index->dir[len-1] == '/')
		index->dir[len-1] = '\0';

	return (MailIndex *) index;
}

static void maildir_index_free(MailIndex *index)
{
	mail_index_close(index);
	i_free(index->dir);
	i_free(index);
}

static int maildir_index_update_flags(MailIndex *index, MailIndexRecord *rec,
				      unsigned int seq, MailFlags flags,
				      int external_change)
{
	MailIndexUpdate *update;
	const char *old_fname, *new_fname;
	const char *old_path, *new_path;

	if (!mail_index_update_flags(index, rec, seq, flags, external_change))
		return FALSE;

	/* we need to update the flags in the file name */
	old_fname = index->lookup_field(index, rec, FIELD_TYPE_LOCATION);
	if (old_fname == NULL) {
                INDEX_MARK_CORRUPTED(index);
		index_set_error(index, "Corrupted index file %s: "
				"Missing location field for record %u",
				index->filepath, rec->uid);
		return FALSE;
	}

	new_fname = maildir_filename_set_flags(old_fname, flags);

	if (strcmp(old_fname, new_fname) != 0) {
		old_path = t_strconcat(index->dir, "/cur/", old_fname, NULL);
		new_path = t_strconcat(index->dir, "/cur/", new_fname, NULL);

		/* minor problem: new_path is overwritten if it exists.. */
		if (rename(old_path, new_path) == -1) {
			index_set_error(index, "maildir flags update: "
					"rename(%s, %s) failed: %s",
					old_path, new_path, strerror(errno));
			return FALSE;
		}

		/* update the filename in index */
		update = index->update_begin(index, rec);
		index->update_field(update, FIELD_TYPE_LOCATION, new_fname, 0);

		if (!index->update_end(update))
			return FALSE;
	}

	return TRUE;
}

static MailIndex maildir_index = {
	mail_index_open,
	mail_index_open_or_create,
	maildir_index_free,
	mail_index_set_lock,
	mail_index_try_lock,
	maildir_index_rebuild,
	mail_index_fsck,
	maildir_index_sync,
	mail_index_get_header,
	mail_index_lookup,
	mail_index_next,
        mail_index_lookup_uid_range,
	mail_index_lookup_field,
	mail_index_get_sequence,
	maildir_open_mail,
	mail_index_expunge,
	maildir_index_update_flags,
	mail_index_append,
	mail_index_update_begin,
	mail_index_update_end,
	mail_index_update_field,
	mail_index_get_last_error,
	mail_index_is_inconsistency_error
};
