/*
 dynarray.c : Dynamically growing arrays

    Copyright (c) 2001-2002 Timo Sirainen

    Permission is hereby granted, free of charge, to any person obtaining
    a copy of this software and associated documentation files (the
    "Software"), to deal in the Software without restriction, including
    without limitation the rights to use, copy, modify, merge, publish,
    distribute, sublicense, and/or sell copies of the Software, and to
    permit persons to whom the Software is furnished to do so, subject to
    the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
    OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
    IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
    CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
    TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#include "lib.h"
#include "dynarray.h"

typedef struct {
	int last;
	void **array;

        Pool pool;
	int size, alloc_size;
	int frozen, frozen_removed_count;
        int destroyed;
} PRealArray;

PArray parray_create(Pool pool, int initial_items)
{
	PRealArray *array;

	array = p_new(pool, PRealArray, 1);
	array->pool = pool;
	array->alloc_size = initial_items;
	array->array = p_new(pool, void *, initial_items);
	return (PArray) array;
}

void parray_destroy(PArray array)
{
	PRealArray *_array = (PRealArray *) array;
	Pool pool;

	if (array == NULL)
                return;

	if (_array->frozen) {
                parray_clear(array);
		_array->destroyed = TRUE;
                return;
	}

        pool = _array->pool;
        p_free(pool, _array->array);
	p_free(pool, _array);
}

int parray_alloc(PArray array, void *data)
{
	PRealArray *_array = (PRealArray *) array;

        return_val_if_fail(data != NULL, -1);

	if (_array->alloc_size == _array->last) {
		/* we'll need to allocate more space - grow the array 1.5x */
		_array->alloc_size += (_array->alloc_size+1)/2;

		/* NOTE: p_realloc() clears the new space so we don't need to */
		_array->array = p_realloc(_array->pool, _array->array,
					  _array->alloc_size * sizeof(void *));
	}

        _array->size++;
	_array->array[_array->last++] = data;
	return _array->last-1;
}

void parray_free(PArray array, int index)
{
	PRealArray *_array = (PRealArray *) array;

        return_if_fail(_array->last > 0);

        _array->size--;

	if (index == _array->last-1) {
                /* removing last element, easy */
		_array->array[--_array->last] = NULL;

                /* make sure `last' always points to non-NULL item */
		while (_array->last > 0 &&
		       _array->array[_array->last-1] == NULL)
                        _array->last--;
	} else if (!_array->frozen) {
		/* move the last item over free'd item */
		_array->last--;
		_array->array[index] = _array->array[_array->size];
		_array->array[_array->last] = NULL;
	} else {
		/* array is frozen, just mark it empty */
                _array->array[index] = NULL;
		_array->frozen_removed_count++;
	}
}

void parray_clear(PArray array)
{
	PRealArray *_array = (PRealArray *) array;

	if (array == NULL)
		return;

	memset(_array->array, 0, sizeof(void *) * _array->last);
        _array->size = _array->last = 0;
}

void parray_foreach(PArray array, PForeachFunc func, void *user_data)
{
	int i;

	if (array == NULL)
                return;

        parray_freeze(array);
	for (i = 0; i < array->last; i++) {
		void *data = parray_index(array, i);

                if (data != NULL)
			func(data, user_data);
	}
        parray_thaw(array);
}

int parray_find(PArray array, void *data)
{
	int i;

	if (array == NULL)
                return -1;

	for (i = 0; i < array->last; i++) {
		if (parray_index(array, i) == data)
                        return i;
	}

        return -1;
}

int parray_find_custom(PArray array, PFindFunc func, void *user_data)
{
	int i;

	if (array == NULL)
                return -1;

	for (i = 0; i < array->last; i++) {
		void *data = parray_index(array, i);

		if (data != NULL && func(data, user_data))
                        return i;
	}

        return -1;
}

void *parray_find_custom_data(PArray array, PFindFunc func, void *user_data)
{
	int index;

	index = parray_find_custom(array, func, user_data);
	return index == -1 ? NULL : parray_index(array, index);
}

void parray_freeze(PArray array)
{
	PRealArray *_array = (PRealArray *) array;

	_array->frozen++;
}

void parray_thaw(PArray array)
{
	PRealArray *_array = (PRealArray *) array;
	int i;

	if (--_array->frozen > 0)
		return;

	if (_array->destroyed) {
		parray_destroy(array);
                return;
	}

	if (_array->frozen_removed_count == 0)
		return;

	if (_array->last == 0) {
                /* everything removed */
	} else {
		/* compress the array */
		for (i = 0; i < _array->last; i++) {
			if (_array->array[i] == NULL) {
                                _array->last--;
				_array->array[i] = _array->array[_array->last];
				_array->array[_array->last] = NULL;

				if (--_array->frozen_removed_count == 0)
                                        break;
			}
		}
	}

        _array->frozen_removed_count = 0;
}

int parray_size(PArray array)
{
	PRealArray *_array = (PRealArray *) array;
        int i, size;

	if (!_array->frozen)
		return _array->last;

        size = 0;
	for (i = 0; i < _array->last; i++) {
		if (parray_index(array, i) != NULL)
                        size++;
	}

        return size;
}
