/* GLIB - Library of useful routines for C programming
 * Copyright (C) 1995-1997  Peter Mattis, Spencer Kimball and Josh MacDonald
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
 * Modified by the GLib Team and others 1997-1999.  See the AUTHORS
 * file for a list of people on the GLib Team.  See the ChangeLog
 * files for a list of changes.  These files are distributed with
 * GLib at ftp://ftp.gtk.org/pub/gtk/. 
 */

#include "lib.h"

struct _Tree {
	int refcount;
	Pool pool;

	TreeNode *root;
	TreeCompareFunc key_compare;
};

struct _TreeNode {
	int balance;      /* height (left) - height (right) */
	TreeNode *left;   /* left subtree */
	TreeNode *right;  /* right subtree */
	void *key;        /* key for this node */
	void *value;      /* value stored at this node */
};

static TreeNode *tree_node_new(Tree *tree, const void *key, const void *value);
static void tree_node_destroy(Tree *tree, TreeNode *node);
static TreeNode *tree_node_insert(Tree *tree, TreeNode *node,
				  TreeCompareFunc compare,
				  const void *key, const void *value,
				  int *inserted);
static TreeNode *tree_node_remove(Tree *tree, TreeNode *node,
				  TreeCompareFunc compare, const void *key);

static int direct_compare(const void *p1, const void *p2)
{
	return p1 < p2 ? -1 : p1 == p2 ? 0 : 1;
}

void tree_ref(Tree *tree)
{
	tree->refcount++;
}

void tree_unref(Tree *tree)
{
	if (--tree->refcount > 0)
		return;

	tree_node_destroy(tree, tree->root);
	pool_unref(tree->pool);
	p_free(tree->pool, tree);
}

Tree *tree_create(Pool pool, TreeCompareFunc key_compare_func)
{
	Tree *tree;

        i_assert(pool != NULL);

	tree = p_new(pool, Tree, 1);
	tree->refcount = 1;
	tree->pool = pool;
	tree->key_compare = key_compare_func != NULL ?
		key_compare_func : direct_compare;

	pool_ref(pool);
        return tree;
}

void tree_destroy(Tree *tree)
{
	if (tree == NULL)
                return;

	tree_unref(tree);
}

void tree_insert(Tree *tree, const void *key, const void *value)
{
	int inserted;

	i_assert(tree != NULL);

	inserted = FALSE;
	tree->root = tree_node_insert(tree, tree->root, tree->key_compare,
				      key, value, &inserted);
}

void tree_remove(Tree *tree, const void *key)
{
	i_assert(tree != NULL);

	tree->root = tree_node_remove(tree, tree->root, tree->key_compare, key);
}

void *tree_lookup(Tree *tree, const void *key)
{
	TreeNode *node;

	i_assert(tree != NULL);

	node = tree_node_search(tree->root, tree->key_compare, key);
	return node == NULL ? NULL : node->value;
}

void tree_traverse(Tree *tree, TreeTraverseFunc traverse_func, void *data)
{
	i_assert(tree != NULL);

	if (tree->root != NULL) {
		tree_ref(tree);
		tree_node_traverse(tree->root, traverse_func, data);
		tree_unref(tree);
	}
}

TreeNode *tree_search(Tree *tree, TreeCompareFunc func, const void *key)
{
	TreeNode *node;

	i_assert(tree != NULL);

	if (tree->root == NULL)
		return NULL;
	if (func == NULL)
		func = tree->key_compare;

	tree_ref(tree);
	node = tree_node_search(tree->root, func, key);
	tree_unref(tree);

	return node;
}

int tree_height(Tree *tree)
{
	i_assert(tree != NULL);

	if (tree->root == NULL)
		return 0;

	return tree_node_height(tree->root);
}

int tree_nodes_count(Tree *tree)
{
	i_assert(tree != NULL);

	if (tree->root == NULL)
		return 0;

	return tree_node_count(tree->root);
}

static TreeNode *tree_node_new(Tree *tree, const void *key, const void *value)
{
	TreeNode *node;

        node = p_new(tree->pool, TreeNode, 1);
	node->key = (void *) key;
	node->value = (void *) value;

	return node;
}

static void tree_node_destroy(Tree *tree, TreeNode *node)
{
	if (node == NULL)
		return;

	tree_node_destroy(tree, node->right);
	tree_node_destroy(tree, node->left);
        p_free(tree->pool, node);
}

static TreeNode *tree_node_rotate_left(TreeNode *node)
{
	TreeNode *left;
	TreeNode *right;
	int a_bal;
	int b_bal;

	left = node->left;
	right = node->right;

	node->right = right->left;
	right->left = node;

	a_bal = node->balance;
	b_bal = right->balance;

	if (b_bal <= 0) {
		if (a_bal >= 1)
			right->balance = b_bal - 1;
		else
			right->balance = a_bal + b_bal - 2;
		node->balance = a_bal - 1;
	} else {
		if (a_bal <= b_bal)
			right->balance = a_bal - 2;
		else
			right->balance = b_bal - 1;
		node->balance = a_bal - b_bal - 1;
	}

	return right;
}

static TreeNode *tree_node_rotate_right(TreeNode *node)
{
	TreeNode *left;
	int a_bal;
	int b_bal;

	left = node->left;

	node->left = left->right;
	left->right = node;

	a_bal = node->balance;
	b_bal = left->balance;

	if (b_bal <= 0) {
		if (b_bal > a_bal)
			left->balance = b_bal + 1;
		else
			left->balance = a_bal + 2;
		node->balance = a_bal - b_bal + 1;
	} else {
		if (a_bal <= -1)
			left->balance = b_bal + 1;
		else
			left->balance = a_bal + b_bal + 2;
		node->balance = a_bal + 1;
	}

	return left;
}
static TreeNode *tree_node_balance(TreeNode *node)
{
	if (node->balance < -1) {
		if (node->left->balance > 0)
			node->left = tree_node_rotate_left(node->left);
		node = tree_node_rotate_right(node);
	} else if (node->balance > 1) {
		if (node->right->balance < 0)
			node->right = tree_node_rotate_right(node->right);
		node = tree_node_rotate_left(node);
	}

	return node;
}

static TreeNode *tree_node_insert(Tree *tree, TreeNode *node,
				  TreeCompareFunc compare,
				  const void *key, const void *value,
				  int *inserted)
{
	int old_balance;
	int cmp;

	if (node == NULL) {
		*inserted = TRUE;
		return tree_node_new(tree, key, value);
	}

	cmp = compare(key, node->key);
	if (cmp == 0) {
		*inserted = FALSE;
		node->value = (void *) value;
		return node;
	}

	if (cmp < 0) {
		if (node->left != NULL) {
			old_balance = node->left->balance;
			node->left = tree_node_insert(tree, node->left, compare,
						      key, value, inserted);

			if ((old_balance != node->left->balance) &&
			    node->left->balance)
				node->balance--;
		} else {
			*inserted = TRUE;
			node->left = tree_node_new(tree, key, value);
			node->balance--;
		}
	} else if (cmp > 0) {
		if (node->right != NULL) {
			old_balance = node->right->balance;
			node->right = tree_node_insert(tree, node->right,
						compare, key, value, inserted);

			if ((old_balance != node->right->balance) &&
			    node->right->balance)
				node->balance++;
		} else {
			*inserted = TRUE;
			node->right = tree_node_new(tree, key, value);
			node->balance++;
		}
	}

	if (*inserted) {
		if ((node->balance < -1) || (node->balance > 1))
			node = tree_node_balance(node);
	}

	return node;
}

static TreeNode *
tree_node_restore_left_balance(TreeNode *node, int old_balance)
{
	if (node->left == NULL)
		node->balance++;
	else if ((node->left->balance != old_balance) &&
		 (node->left->balance == 0))
		node->balance++;

	if (node->balance > 1)
		return tree_node_balance(node);
	return node;
}

static TreeNode *
tree_node_restore_right_balance(TreeNode *node, int old_balance)
{
	if (node->right == NULL)
		node->balance--;
	else if ((node->right->balance != old_balance) &&
		 (node->right->balance == 0))
		node->balance--;

	if (node->balance < -1)
		return tree_node_balance(node);
	return node;
}

static TreeNode *
tree_node_remove_leftmost(TreeNode  *node, TreeNode **leftmost)
{
	int old_balance;

	if (node->left == NULL) {
		*leftmost = node;
		return node->right;
	}

	old_balance = node->left->balance;
	node->left = tree_node_remove_leftmost(node->left, leftmost);
	return tree_node_restore_left_balance(node, old_balance);
}

static TreeNode *tree_node_remove(Tree *tree, TreeNode *node,
				  TreeCompareFunc compare, const void *key)
{
	TreeNode *new_root;
	int old_balance;
	int cmp;

	if (node == NULL)
		return NULL;

	cmp = compare(key, node->key);
	if (cmp == 0) {
		TreeNode *old_node;

                old_node = node;
		if (node->right == NULL) {
			node = node->left;
		} else {
			old_balance = node->right->balance;
			node->right = tree_node_remove_leftmost(node->right,
								&new_root);
			new_root->left = node->left;
			new_root->right = node->right;
			new_root->balance = node->balance;
			node = tree_node_restore_right_balance(new_root,
							       old_balance);
		}

                p_free(tree->pool, old_node);
	} else if (cmp < 0) {
		if (node->left) {
			old_balance = node->left->balance;
			node->left = tree_node_remove(tree, node->left,
						      compare, key);
			node = tree_node_restore_left_balance(node,
							      old_balance);
		}
	} else if (cmp > 0) {
		if (node->right) {
			old_balance = node->right->balance;
			node->right = tree_node_remove(tree, node->right,
						       compare, key);
			node = tree_node_restore_right_balance(node,
							       old_balance);
		}
	}

	return node;
}

TreeNode *tree_node_search(TreeNode *node, TreeCompareFunc compare,
			   const void *key)
{
	int dir;

	if (node == NULL)
		return NULL;

	do {
		dir = compare(key, node->key);
		if (dir == 0)
			return node;

		if (dir < 0)
			node = node->left;
		else
			node = node->right;
	} while (node != NULL);

	return NULL;
}

int tree_node_traverse(TreeNode *node, TreeTraverseFunc traverse_func,
		       void *data)
{
	if (node->left != NULL) {
		if (!tree_node_traverse(node->left, traverse_func, data))
			return FALSE;
	}

	if (!traverse_func(node->key, node->value, data))
		return FALSE;

	if (node->right != NULL) {
		if (!tree_node_traverse(node->right, traverse_func, data))
			return FALSE;
	}

	return TRUE;
}

int tree_node_count(TreeNode *node)
{
	int count;

	count = 1;
	if (node->left)
		count += tree_node_count(node->left);
	if (node->right)
		count += tree_node_count(node->right);

	return count;
}

int tree_node_height(TreeNode *node)
{
	int left_height;
	int right_height;

	if (node) {
		left_height = 0;
		right_height = 0;

		if (node->left)
			left_height = tree_node_height(node->left);

		if (node->right)
			right_height = tree_node_height(node->right);

		return MAX(left_height, right_height) + 1;
	}

	return 0;
}
